#!/bin/bash

# parameter 1 is the path to a file generated by ioreg

thefile=$1

matchre1='+-o'

elcapre='.*"IOPCIExpressLinkCapabilities" = ([0-9]+)'
elsatre='.*"IOPCIExpressLinkStatus" = ([0-9]+)'
pcicapre='.*"IOPCIExpressCapabilities" = ([0-9]+)'
namere='.*"name" = .+'
compatiblere='.*"compatible" = .+'
vendoridre='(.*"vendor-id" = )<(..)(..)....>'
deviceidre='(.*"device-id" = )<(..)(..)....>'
subsystemidre='(.*"subsystem-id" = )<(..)(..)....>'
subsystemvendoridre='(.*"subsystem-vendor-id" = )<(..)(..)....>'

theobject=''
elcap=''
elsat=''
pcicap=''
elcapline=''
elsatline=''
pcicapline=''
nameline=''
compatibleline=''
vendoridline=''
deviceidline=''
subsystemidline=''
subsystemvendoridline=''


echo $thefile
(cat $thefile ; echo '+-o') | \
while IFS='\n' read -r line || [[ -n $line ]]; do
	if [[ $line =~ "$matchre1" ]]; then
		if [[ -n "$elcap$elsat$pcicap" ]]; then
			echo "$theobject"
			if [[ -n "$nameline" ]]; then echo "$nameline"; fi
			if [[ -n "$compatibleline" ]]; then echo "$compatibleline"; fi
			if [[ -n "$vendoridline" ]]; then echo "$vendoridline"; fi
			if [[ -n "$deviceidline" ]]; then echo "$deviceidline"; fi
			if [[ -n "$subsystemidline" ]]; then echo "$subsystemidline"; fi
			if [[ -n "$subsystemvendoridline" ]]; then echo "$subsystemvendoridline"; fi

			echo -n "$pcicapline ("$(printf "0x%04x" $pcicap)"):"
			# $000F Capability Version
			# $00F0 Device/Port Type
			# $0100 Slot Implemented
			# $3E00 Interrupt Message Number
			# $4000 Undefined
			# $8000 Reserved
			echo -n " Version:"$(($pcicap & 0x000F))
			case $((($pcicap & 0x00F0) >> 4)) in
				 0) echo -n " (PCI Express Endpoint)" ;;
				 1) echo -n " (Legacy PCI Express Endpoint)" ;;
				 4) echo -n " (Root Port of PCI Express Root Complex)" ;;
				 5) echo -n " (Upstream Port of PCI Express Switch)" ;;
				 6) echo -n " (Downstream Port of PCI Express Switch)" ;;
				 7) echo -n " (PCI Express to PCI/PCI-X Bridge)" ;;
				 8) echo -n " (PCI/PCI-X to PCI Express Bridge)" ;;
				 9) echo -n " (Root Complex Integrated Endpoint)" ;;
				10) echo -n " (Root Complex Event Collector)" ;;
				 *) echo -n " (Reserved Type "$((($pcicap & 0x00F0)>>4))")" ;;
			esac
			if [[ $(($pcicap & 0x0100)) -ne 0 ]]; then echo -n " SlotImplemented"; fi
			echo " Interrupt:"$((($pcicap & 0x3E00) >> 9))

			echo -n "$elcapline ("$(printf "0x%08x" $elcap)"):"
			# $0000000F Max Link Speed
			# $000003F0 Maximum Link Width
			# $00000C00 ASPM Support
			# $00007000 L0s Exit Latency
			# $00038000 L1 Exit Latency
			# $00040000 Clock Power Management
			# $00080000 Surprise Down Error Reporting Capable
			# $00100000 Data Link Layer Link Active Reporting Capable
			# $00200000 Link Bandwidth Notification Capability
			# $00400000 ASPM Optionality Compliance
			# $00800000 Reserved
			# $FF000000 Port Number
			case $(($elcap & 0x000F)) in
				1) echo -n " 2.5 GT/s" ;;
				2) echo -n " 5.0 GT/s" ;;
				3) echo -n " 8.0 GT/s" ;;
				*) echo -n " ? GT/s" ;;
			esac
			echo -n " x"$((($elcap & 0x000003F0) >> 4))
			case $((($elcap & 0x00000C00) >> 10)) in
				0) echo -n " NoASPM" ;;
				1) echo -n " L0sSupported" ;;
				2) echo -n " L1Supported" ;;
				3) echo -n " L0sAndL1Supported" ;;
			esac
			case $((($elcap & 0x00007000) >> 12)) in
				0) echo -n " L0s:[0,64)ns" ;;
				1) echo -n " L0s:[64,128)ns" ;;
				2) echo -n " L0s:[128,256)ns" ;;
				3) echo -n " L0s:[256,512)ns" ;;
				4) echo -n " L0s:[512,1000)ns" ;;
				5) echo -n " L0s:[1,2)µs" ;;
				6) echo -n " L0s:[2,4]µs" ;;
				7) echo -n " L0s:(4,∞)µs" ;;
			esac
			case $((($elcap & 0x00038000) >> 15)) in
				0) echo -n " L1:[0,1)µs" ;;
				1) echo -n " L1:[1,2)µs" ;;
				2) echo -n " L1:[2,4)µs" ;;
				3) echo -n " L1:[4,8)µs" ;;
				4) echo -n " L1:[8,16)µs" ;;
				5) echo -n " L1:[16,32)µs" ;;
				6) echo -n " L1:[32,64]µs" ;;
				7) echo -n " L1:(64,∞)µs" ;;
			esac
			if [[ $(($elcap & 0x00040000)) -ne 0 ]]; then echo -n " AllowReferenceClockRemoval"; fi
			if [[ $(($elcap & 0x00080000)) -ne 0 ]]; then echo -n " SurpriseDownErrorReporting"; fi
			if [[ $(($elcap & 0x00100000)) -ne 0 ]]; then echo -n " ActiveReporting"; fi
			if [[ $(($elcap & 0x00200000)) -ne 0 ]]; then echo -n " BandwidthNotifications"; fi
			if [[ $(($elcap & 0x00400000)) -ne 0 ]]; then echo -n " ASPM"; fi
			echo " Port:$(($elcap >> 24))"
			
			echo -n "$elsatline ("$(printf "0x%04x" $elsat)"):"
			# $000F Current Link Speed
			# $03F0 Negotiated Link Width
			# $0400 Undefined
			# $0800 Link Training
			# $1000 Slot Clock Configuration
			# $2000 Data Link Layer Link Active
			# $4000 Link Bandwidth Management Status
			# $8000 Link Autonomous Bandwidth Status
			case $(($elsat & 0x000F)) in
				1) echo -n " 2.5 GT/s" ;;
				2) echo -n " 5.0 GT/s" ;;
				3) echo -n " 8.0 GT/s" ;;
				*) echo -n " ? GT/s" ;;
			esac
			echo -n " x"$((($elsat & 0x03F0) >> 4))
			if [[ $(($elsat & 0x0400)) -ne 0 ]]; then echo -n " Undefined"; fi
			if [[ $(($elsat & 0x0800)) -ne 0 ]]; then echo -n " Training"; fi
			if [[ $(($elsat & 0x1000)) -ne 0 ]]; then echo -n " SlotClocked"; fi
			if [[ $(($elsat & 0x2000)) -ne 0 ]]; then echo -n " Active"; fi
			if [[ $(($elsat & 0x4000)) -ne 0 ]]; then echo -n " Managed"; fi
			if [[ $(($elsat & 0x8000)) -ne 0 ]]; then echo -n " Changed"; fi
			echo

		fi
		theobject=$line
		elcap=''
		elsat=''
		pcicap=''
		elcapline=''
		elsatline=''
		pcicapline=''
		nameline=''
		compatibleline=''
		vendoridline=''
		deviceidline=''
		subsystemidline=''
		subsystemvendoridline=''
	elif [[ $line =~ $elcapre ]]; then
		elcap=${BASH_REMATCH[1]}
		elcapline=$line
	elif [[ $line =~ $elsatre ]]; then
		elsat=${BASH_REMATCH[1]}
		elsatline=$line
	elif [[ $line =~ $pcicapre ]]; then
		pcicap=${BASH_REMATCH[1]}
		pcicapline=$line
	elif [[ $line =~ $namere ]]; then
		nameline=$line
	elif [[ $line =~ $compatiblere ]]; then
		compatibleline=$line
	elif [[ $line =~ $vendoridre ]]; then
		vendoridline=${BASH_REMATCH[1]}${BASH_REMATCH[3]}${BASH_REMATCH[2]}
	elif [[ $line =~ $deviceidre ]]; then
		deviceidline=${BASH_REMATCH[1]}${BASH_REMATCH[3]}${BASH_REMATCH[2]}
	elif [[ $line =~ $subsystemidre ]]; then
		subsystemidline=${BASH_REMATCH[1]}${BASH_REMATCH[3]}${BASH_REMATCH[2]}
	elif [[ $line =~ $subsystemvendoridre ]]; then
		subsystemvendoridline=${BASH_REMATCH[1]}${BASH_REMATCH[3]}${BASH_REMATCH[2]}
	fi
done
